"""Decorate functions with contracts."""
import abc
import functools
import inspect
import os
import reprlib
from typing import Callable, MutableMapping, Any, Optional, Set, List, Type, Dict, \
    Tuple, Mapping, Iterable  # pylint: disable=unused-import

import icontract.represent


class ViolationError(AssertionError):
    """Indicate a violation of a contract."""

    pass


# Default representation instance.
#
# The limits are set way higher than reprlib.aRepr since the default reprlib limits are not suitable for
# the production systems.
aRepr = reprlib.Repr()  # pylint: disable=invalid-name
aRepr.maxdict = 50
aRepr.maxlist = 50
aRepr.maxtuple = 50
aRepr.maxset = 50
aRepr.maxfrozenset = 50
aRepr.maxdeque = 50
aRepr.maxarray = 50
aRepr.maxstring = 256
aRepr.maxother = 256

# SLOW provides a unified environment variable (ICONTRACT_SLOW) to enable the contracts which are slow to execute.
#
# Use SLOW to mark any contracts that are even too slow to make it to the normal (__debug__) execution of
# the interpreted program.
#
# Contracts marked with SLOW are also disabled if the interpreter is run in optimized mode (``-O`` or ``-OO``).
SLOW = __debug__ and os.environ.get("ICONTRACT_SLOW", "") != ""


class _Contract:
    """Represent a contract to be enforced as a precondition, postcondition or as an invariant."""

    def __init__(self,
                 condition: Callable[..., bool],
                 description: Optional[str] = None,
                 repr_args: Optional[Callable[..., str]] = None,
                 a_repr: Optional[reprlib.Repr] = None) -> None:
        """
        Initialize.

        :param condition: condition predicate
        :param description: textual description of the contract
        :param repr_args:
            function to represent arguments in the message on a failed condition. The repr_func needs to take the
            same arguments as the condition function.

            If not specified, all the involved values are represented by re-traversing the AST.
        :param a_repr:
            representation instance that defines how the values are represented.

            If ``repr_args`` is specified, ``repr_instance`` should be None.
            If no ``repr_args`` is specified, the default ``aRepr`` is used.

        """
        # pylint: disable=too-many-arguments
        if repr_args is not None and a_repr is not None:
            raise ValueError("Expected no repr_instance if repr_args is given.")

        self.condition = condition

        self.condition_args = list(inspect.signature(condition).parameters.keys())  # type: List[str]
        self.condition_arg_set = set(self.condition_args)  # type: Set[str]
        self.condition_as_text = icontract.represent.condition_as_text(condition=condition)

        self.description = description

        self._repr_func = repr_args
        if repr_args is not None:
            got = list(inspect.signature(repr_args).parameters.keys())

            if got != self.condition_args:
                raise ValueError("Unexpected argument(s) of repr_args. Expected {}, got {}".format(
                    self.condition_args, got))

        self._a_repr = a_repr if a_repr is not None else aRepr

    def __repr__(self) -> str:
        return "{} at {:x} on {}".format(type(self), id(self), self.condition_as_text)


def _generate_message(contract: _Contract, condition_kwargs: Mapping[str, Any]) -> str:
    """Generate the message upon contract violation."""
    # pylint: disable=protected-access
    parts = []  # type: List[str]

    if contract.description is not None:
        parts.append("{}: ".format(contract.description))

    parts.append(contract.condition_as_text)

    if contract._repr_func:
        parts.append(': ')
        parts.append(contract._repr_func(**condition_kwargs))
    else:
        repr_values = icontract.represent.repr_values(
            condition=contract.condition, condition_kwargs=condition_kwargs, a_repr=contract._a_repr)

        if len(repr_values) == 1:
            parts.append(': ')
            parts.append(repr_values[0])
        else:
            parts.append(':\n')
            parts.append('\n'.join(repr_values))

    msg = "".join(parts)
    return msg


def _assert_precondition(contract: _Contract, param_names: List[str], kwdefaults: Dict[str, Any], args: Tuple[Any, ...],
                         kwargs: Dict[str, Any]) -> None:
    """Assert that the contract holds as a precondition to the function ``func``."""
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    condition_kwargs = dict()  # type: MutableMapping[str, Any]

    # Set the default argument values as condition parameters.
    for param_name, param_value in kwdefaults.items():
        if param_name in contract.condition_arg_set:
            condition_kwargs[param_name] = param_value

    # Override the defaults with the values actually suplied to the function.
    for i, func_arg in enumerate(args):
        if param_names[i] in contract.condition_arg_set:
            condition_kwargs[param_names[i]] = func_arg

    for key, val in kwargs.items():
        if key in contract.condition_arg_set:
            condition_kwargs[key] = val

    # Check that all arguments to the condition function have been set.
    for arg_name in contract.condition_args:
        if arg_name not in condition_kwargs:
            raise TypeError(("The argument of the contract condition has not been set: {}. "
                             "Does the function define it?").format(arg_name))

    check = contract.condition(**condition_kwargs)

    if not check:
        msg = _generate_message(contract=contract, condition_kwargs=condition_kwargs)
        raise ViolationError(msg)


def _assert_invariant(contract: _Contract, instance: Any) -> None:
    """Assert that the contract holds as a class invariant given the instance of the class."""
    check = contract.condition(self=instance)

    if not check:
        msg = _generate_message(contract=contract, condition_kwargs={"self": instance})
        raise ViolationError(msg)


def _assert_postcondition(contract: _Contract, param_names: List[str], kwdefaults: Dict[str, Any],
                          args: Tuple[Any, ...], kwargs: Dict[str, Any], result: Any) -> None:
    """Assert that the contract holds as a postcondition given the arguments and the result of a function."""
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    condition_kwargs = dict()  # type: MutableMapping[str, Any]

    # Set the default argument values as condition parameters.
    for param_name, param_value in kwdefaults.items():
        if param_name in contract.condition_arg_set:
            condition_kwargs[param_name] = param_value

    # Override the default argument values with the values actually supplied to the function.
    for i, func_arg in enumerate(args):
        if param_names[i] in contract.condition_arg_set:
            condition_kwargs[param_names[i]] = func_arg

    # Collect the keyword arguments
    for key, val in kwargs.items():
        if key in contract.condition_arg_set:
            condition_kwargs[key] = val

    # Add the special ``result`` argument
    if "result" in contract.condition_arg_set:
        condition_kwargs["result"] = result

    # Check that all arguments to the condition function have been set.
    for arg_name in contract.condition_args:
        if arg_name not in condition_kwargs:
            raise TypeError(("The argument of the contract condition has not been set: {}. "
                             "Does the function define it?").format(arg_name))

    check = contract.condition(**condition_kwargs)

    if not check:
        msg = _generate_message(contract=contract, condition_kwargs=condition_kwargs)
        raise ViolationError(msg)


def _decorate_with_checker(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorate the function with a checker that verifies the preconditions and postconditions."""
    assert not hasattr(func, "__preconditions__"), \
        "Expected func to have no list of preconditions (there should be only a single contract checker per function)."

    assert not hasattr(func, "__postconditions__"), \
        "Expected func to have no list of postconditions (there should be only a single contract checker per function)."

    sign = inspect.signature(func)
    param_names = list(sign.parameters.keys())

    # Determine the default argument values.
    kwdefaults = dict()  # type: Dict[str, Any]

    # Add to the defaults all the values that are needed by the contracts.
    for param in sign.parameters.values():
        if param.default != inspect.Parameter.empty:
            kwdefaults[param.name] = param.default

    def wrapper(*args, **kwargs):
        """Wrap func by checking the preconditions and postconditions."""
        preconditions = getattr(wrapper, "__preconditions__")
        postconditions = getattr(wrapper, "__postconditions__")

        # Assert the preconditions in groups. This is necessary to implement "require else" logic when a class
        # weakens the preconditions of its base class.
        violation_err = None  # type: Optional[ViolationError]
        for group in preconditions:
            violation_err = None
            try:
                for contract in group:
                    _assert_precondition(
                        contract=contract, param_names=param_names, kwdefaults=kwdefaults, args=args, kwargs=kwargs)
                break
            except ViolationError as err:
                violation_err = err

        if violation_err is not None:
            raise violation_err  # pylint: disable=raising-bad-type

        # Execute the wrapped function
        result = func(*args, **kwargs)

        # Assert the postconditions as a conjunction
        for contract in postconditions:
            _assert_postcondition(
                contract=contract,
                param_names=param_names,
                kwdefaults=kwdefaults,
                args=args,
                kwargs=kwargs,
                result=result)

        return result

    # Copy __doc__ and other properties so that doctests can run
    functools.update_wrapper(wrapper=wrapper, wrapped=func)

    assert not hasattr(wrapper, "__preconditions__"), "Expected no preconditions set on a pristine contract checker."
    assert not hasattr(wrapper, "__postconditions__"), "Expected no postconditions set on a pristine contract checker."

    # Precondition is a list of condition groups (i.e. disjunctive normal form):
    # each group consists of AND'ed preconditions, while the groups are OR'ed.
    #
    # This is necessary in order to implement "require else" logic when a class weakens the preconditions of
    # its base class.
    setattr(wrapper, "__preconditions__", [[]])

    setattr(wrapper, "__postconditions__", [])

    return wrapper


def _unwind_decorator_stack(func: Callable[..., Any]) -> Iterable[Callable[..., Any]]:
    """
    Iterate through the stack of decorated functions and return the original function.

    Assume that all decorators used functools.update_wrapper.
    """
    while hasattr(func, "__wrapped__"):
        yield func

        func = getattr(func, "__wrapped__")

    yield func


def _find_checker(func: Callable[..., Any]) -> Optional[Callable[..., Any]]:
    """Iterate through the decorator stack till we find the contract checker."""
    contract_checker = None  # type: Optional[Callable[..., Any]]
    for a_wrapper in _unwind_decorator_stack(func):
        if hasattr(a_wrapper, "__preconditions__") or hasattr(a_wrapper, "__postconditions__"):
            contract_checker = a_wrapper

    return contract_checker


class pre:  # pylint: disable=invalid-name
    """
    Decorate a function with a precondition.

    The arguments of the precondition are expected to be a subset of the arguments of the wrapped function.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self,
                 condition: Callable[..., bool],
                 description: Optional[str] = None,
                 repr_args: Optional[Callable[..., str]] = None,
                 a_repr: Optional[reprlib.Repr] = None,
                 enabled: bool = __debug__) -> None:
        """
        Initialize.

        :param condition: precondition predicate
        :param description: textual description of the precondition
        :param repr_args:
            function to represent arguments in the message on a failed precondition. The repr_func needs to take the
            same arguments as the condition function.

            If not specified, all the involved values are represented by re-traversing the AST.
        :param a_repr:
            representation instance that defines how the values are represented.

            If ``repr_args`` is specified, ``repr_instance`` should be None.
            If no ``repr_args`` is specified, the default ``aRepr`` is used.
        :param enabled:
            The decorator is applied only if this argument is set.

            Otherwise, the condition check is disabled and there is no run-time overhead.

            The default is to always check the condition unless the interpreter runs in optimized mode (``-O`` or
            ``-OO``).

        """
        # pylint: disable=too-many-arguments
        self.enabled = enabled
        self._contract = None  # type: Optional[_Contract]

        if not enabled:
            return

        self._contract = _Contract(condition=condition, description=description, repr_args=repr_args, a_repr=a_repr)

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Add the precondition to the list of preconditions of the function ``func``.

        The function ``func`` is decorated with a contract checker if there is no contract checker in
        the decorator stack.

        :param func: function to be wrapped
        :return: contract checker around ``func`` if no contract checker on the decorator stack, or ``func`` otherwise
        """
        if not self.enabled:
            return func

        # Find a contract checker
        contract_checker = _find_checker(func=func)

        if contract_checker is not None:
            # Do not add an additional wrapper since the function has been already wrapped with a contract checker
            result = func
        else:
            # Wrap the function with a contract checker
            contract_checker = _decorate_with_checker(func=func)
            result = contract_checker

        # Add the precondition to the list of preconditions stored at the checker
        assert hasattr(contract_checker, "__preconditions__")
        preconditions = getattr(contract_checker, "__preconditions__")
        assert isinstance(preconditions, list)
        assert len(preconditions) == 1, \
            "A single group of preconditions expected when wrapping with a contract checker. " \
            "The preconditions are merged only in the DBC metaclass."

        assert isinstance(preconditions[0], list)
        preconditions[0].append(self._contract)

        return result


class post:  # pylint: disable=invalid-name
    """
    Decorate a function with a postcondition.

    The arguments of the postcondition are expected to be a subset of the arguments of the wrapped function.
    Additionally, the argument "result" is reserved for the result of the wrapped function. The wrapped function must
    not have "result" among its arguments.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self,
                 condition: Callable[..., bool],
                 description: Optional[str] = None,
                 repr_args: Optional[Callable[..., str]] = None,
                 a_repr: Optional[reprlib.Repr] = None,
                 enabled: bool = __debug__) -> None:
        """
        Initialize.

        :param condition: postcondition predicate
        :param description: textual description of the postcondition
        :param repr_args:
            function to represent arguments in the message on a failed postcondition. The repr_func needs to take the
            same arguments as the condition function.

            If not specified, all the involved values are represented by re-traversing the AST.
        :param a_repr:
            representation instance that defines how the values are represented.

            If ``repr_args`` is specified, ``repr_instance`` should be None.
            If no ``repr_args`` is specified, the default ``reprlib.aRepr`` is used.
        :param enabled:
            The decorator is applied only if this argument is set.

            Otherwise, the condition check is disabled and there is no run-time overhead.

            The default is to always check the condition unless the interpreter runs in optimized mode (``-O`` or
            ``-OO``).

        """
        # pylint: disable=too-many-arguments
        self.enabled = enabled
        self._contract = None  # type: Optional[_Contract]

        if not enabled:
            return

        self._contract = _Contract(condition=condition, description=description, repr_args=repr_args, a_repr=a_repr)

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Add the postcondition to the list of postconditions of the function ``func``.

        The function ``func`` is decorated with a contract checker if there is no contract checker in
        the decorator stack.

        :param func: function to be wrapped
        :return: contract checker around ``func`` if no contract checker on the decorator stack, or ``func`` otherwise
        """
        if not self.enabled:
            return func

        # Find a contract checker
        contract_checker = _find_checker(func=func)

        if contract_checker is not None:
            # Do not add an additional wrapper since the function has been already wrapped with a contract checker
            result = func
        else:
            # Wrap the function with a contract checker
            contract_checker = _decorate_with_checker(func=func)
            result = contract_checker

        # Add the precondition to the list of preconditions stored at the checker
        assert hasattr(contract_checker, "__postconditions__")
        assert isinstance(getattr(contract_checker, "__postconditions__"), list)
        getattr(contract_checker, "__postconditions__").append(self._contract)

        return result


def _find_self(param_names: List[str], args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Any:
    """Find the instance of ``self`` in the arguments."""
    instance_i = param_names.index("self")
    if instance_i < len(args):
        instance = args[instance_i]
    else:
        instance = kwargs["self"]

    return instance


def _decorate_with_invariants(cls: Type, func: Callable[..., Any], is_init: bool) -> Callable[..., Any]:
    """
    Decorate the function ``func`` of the class ``cls`` with invariant checks.

    :param cls: class to be decorated
    :param func: function to be wrapped
    :param is_init: True if the ``func`` is __init__
    :return: function wrapped with invariant checks
    """
    sign = inspect.signature(func)
    param_names = list(sign.parameters.keys())

    if is_init:

        def wrapper(*args, **kwargs):
            """Wrap __init__ method of a class by checking the invariants *after* the invocation."""
            result = func(*args, **kwargs)
            instance = _find_self(param_names=param_names, args=args, kwargs=kwargs)

            for contract in cls.__invariants__:
                _assert_invariant(contract=contract, instance=instance)

            return result
    else:

        def wrapper(*args, **kwargs):
            """Wrap a function of a class by checking the invariants *before* and *after* the invocation."""
            instance = _find_self(param_names=param_names, args=args, kwargs=kwargs)

            for contract in cls.__invariants__:
                _assert_invariant(contract=contract, instance=instance)

            result = func(*args, **kwargs)

            for contract in cls.__invariants__:
                _assert_invariant(contract=contract, instance=instance)

            return result

    functools.update_wrapper(wrapper=wrapper, wrapped=func)

    setattr(wrapper, "__is_invariant_check__", True)

    return wrapper


def _add_invariant_checks(cls: Type) -> None:
    """Decorate each of the class functions with invariant checks if not already decorated."""
    for name, value in [(name, getattr(cls, name)) for name in dir(cls)]:
        if not inspect.ismethod(value) and not inspect.isfunction(value):
            continue

        # Ignore class methods
        if getattr(value, "__self__", None) is cls:
            continue

        # Ignore __repr__ to avoid endless loops when generating the error message on invariant breach.
        if name == "__repr__":
            continue

        func = value
        for a_decorator in _unwind_decorator_stack(func=func):
            if getattr(a_decorator, "__is_invariant_check__", False):
                continue

        if name == "__init__":
            wrapper = _decorate_with_invariants(cls=cls, func=func, is_init=True)
            setattr(cls, name, wrapper)

        elif not name.startswith("_") or (name.startswith("__") and name.endswith("__")):
            wrapper = _decorate_with_invariants(cls=cls, func=func, is_init=False)
            setattr(cls, name, wrapper)

        elif name.startswith("_"):
            # It is a private or a protected method or function, do not enforce any pre and postconditions.
            pass

        else:
            raise NotImplementedError("Unhandled method or function of class {}: {}".format(cls.__name__, name))


def _identity_decorator(cls: Type) -> Type:
    """Do not decorate the class at all."""
    return cls


def inv(condition: Callable[..., bool],
        description: Optional[str] = None,
        repr_args: Optional[Callable[..., str]] = None,
        a_repr: Optional[reprlib.Repr] = None,
        enabled: bool = __debug__) -> Callable[[Type], Type]:
    """
    Create a class decorator to establish the invariant on all the public methods.

    Class method as well as "private" (prefix ``__``) and "protected" methods (prefix ``_``) may violate the invariant.
    Note that all magic methods (prefix ``__`` and suffix ``__``) are considered public and hence also need to establish
    the invariant. To avoid endless loops when generating the error message on an invariant breach, the method
    ``__repr__`` is deliberately exempt from observing the invariant.

    The invariant is checked *before* and *after* the method invocation.

    :param condition: invariant predicate
    :param description: textual description of the invariant
    :param repr_args:
            function to represent arguments in the message on a breached invariant. The repr_func takes only ``self``
            as a single argument.

            If not specified, all the involved values are represented by re-traversing the AST.

    :param a_repr:
            representation instance that defines how the values are represented.

            If ``repr_args`` is specified, ``repr_instance`` should be None.
            If no ``repr_args`` is specified, the default ``icontract.aRepr`` is used.
    :param enabled:
            The decorator is applied only if this argument is set.

            Otherwise, the condition check is disabled and there is no run-time overhead.

            The default is to always check the condition unless the interpreter runs in optimized mode (``-O`` or
            ``-OO``).

    :return: class decorator

    """
    if not enabled:
        return _identity_decorator

    parameter_names = sorted(inspect.signature(condition).parameters.keys())
    if parameter_names != ["self"]:
        raise ValueError(
            "Expected a condition function with a single argument 'self', but got: {}".format(parameter_names))

    contract = _Contract(condition=condition, description=description, repr_args=repr_args, a_repr=a_repr)

    def decorator(cls: Type) -> Type:
        """
        Decorate each of the public methods with the invariant.

        Go through the decorator stack of each function and search for a contract checker. If there is one,
        add the invariant to the checker's invariants. If there is no checker in the stack, wrap the function with a
        contract checker.
        """
        if not hasattr(cls, "__invariants__"):
            invariants = []  # type: List[_Contract]
            setattr(cls, "__invariants__", invariants)
        else:
            invariants = getattr(cls, "__invariants__")
            assert isinstance(invariants, list), \
                "Expected invariants of class {} to be a list, but got: {}".format(cls, type(invariants))

        invariants.append(contract)

        _add_invariant_checks(cls=cls)

        return cls

    return decorator


class DBCMeta(abc.ABCMeta):
    """
    Define a meta class that allows inheritance of the contracts.

    The preconditions are weakned ("require else"), while postconditions ("ensure then") and invariants are
    strengthened according to the inheritance rules of the design-by-contract.
    """

    def __new__(mcs, name, bases, namespace):
        """Create a class with inherited preconditions, postconditions and invariants."""
        cls = super().__new__(mcs, name, bases, namespace)

        if hasattr(cls, "__invariants__"):
            _add_invariant_checks(cls=cls)

        for key, value in namespace.items():
            for base in bases:
                if hasattr(base, key):
                    # Ignore non-functions
                    if not inspect.ismethod(value) and not inspect.isfunction(value):
                        continue

                    func = value

                    # Check if there is a checker function in the base class
                    base_func = getattr(base, key)
                    base_contract_checker = _find_checker(func=base_func)

                    # Ignore functions which don't have preconditions or postconditions
                    if base_contract_checker is None:
                        continue

                    # Create a contract checker if it has not been defined already
                    contract_checker = _find_checker(func=func)

                    if contract_checker is None:
                        # Inherit the preconditions and postconditions from the base function
                        contract_checker = _decorate_with_checker(func=func)
                        contract_checker.__preconditions__ = base_contract_checker.__preconditions__[:]
                        contract_checker.__postconditions__ = base_contract_checker.__postconditions__[:]
                        setattr(cls, key, contract_checker)
                    else:
                        # Merge the contracts of the base function and this function
                        contract_checker.__preconditions__.extend(base_contract_checker.__preconditions__)
                        contract_checker.__postconditions__.extend(base_contract_checker.__postconditions__)

        return cls


class DBC(metaclass=DBCMeta):
    """Provide a standard way to create a class which can inherit the contracts."""

    pass



"""Generate and visualie the AST as a graph."""
import ast
import pathlib
import subprocess
import tempfile
from typing import List, BinaryIO, cast  # pylint: disable=unused-import

# We do not want to compute the coverage since visualizing AST graphs is used only for debugging.
# pragma: no cover


class Graph:
    """
    Generate DOT graph of an abstract syntax tree (AST).

    We use ``Graph`` to visualize ASTs during debugging.

    """

    def __init__(self) -> None:
        """Initialize."""
        self._nodes = []  # type: List[str]
        self._edges = []  # type: List[str]
        self._init = False

    def visit(self, root: ast.AST) -> None:
        """
        Walk recursively starting from the node and generate the tree.

        :param root: root of the AST
        :return:
        """
        # pylint: disable=too-many-nested-blocks
        self._init = True
        self._edges = []  # type: List[str]
        self._nodes = []  # type: List[str]

        stack = [root]
        while stack:
            node = stack.pop()

            label = ["<b>{}<br/>{:x}</b>".format(type(node).__name__, id(node))]

            for name, field in ast.iter_fields(node):
                if isinstance(field, ast.AST):
                    self._edges.append("node_{} -> node_{} [label={}];".format(id(node), id(field), name))
                    stack.append(field)
                elif isinstance(field, list):
                    if len(field) == 0:
                        label.append("{}: []".format(name))
                    else:
                        if isinstance(field[0], ast.AST):
                            for i, item in enumerate(field):
                                self._edges.append("node_{} -> node_{} [label=<{}[{}]>];".format(
                                    id(node), id(item), name, i))
                            stack.extend(field)

                        else:
                            label.append('{}: {}'.format(name, field))
                else:
                    label.append('{}: {}'.format(name, field))

            self._nodes.append("node_{} [label=<{}>];".format(id(node), "<br/>".join(label)))

    def dot(self) -> str:
        """
        Generate the dot file.

        :return: dot file as a string
        """
        return '\n'.join(['digraph G {'] + self._nodes + self._edges + ['}', ''])

    def show(self) -> None:
        """Call ``graphviz`` and ``display`` to visualize the graph."""
        with tempfile.NamedTemporaryFile() as tmp_in, tempfile.NamedTemporaryFile() as tmp_out:
            tmp_in.file.write(self.dot().encode())  # type: ignore
            tmp_in.file.flush()  # type: ignore
            subprocess.check_call(['dot', tmp_in.name, '-Tsvg', '-o', tmp_out.name])
            subprocess.check_call(['display', tmp_out.name])

    def save_as_svg(self, path: pathlib.Path) -> None:
        """
        Convert the graph from DOT to SVG with ``graphviz``.

        :param path: where the SVG should be stored
        :return:
        """
        with tempfile.NamedTemporaryFile() as tmp_in:
            tmp_in.file.write(self.dot().encode())  # type: ignore
            tmp_in.file.flush()  # type: ignore
            subprocess.check_call(['dot', tmp_in.name, '-Tsvg', '-o', path.as_posix()])



"""Handle recomputation of values of a function given its abstract syntax tree and function frame."""

import ast
import builtins
import functools
from typing import Any, Mapping, Dict, List, Optional, Union, Tuple, Set, Callable  # pylint: disable=unused-import


class Visitor(ast.NodeVisitor):
    """Traverse the abstract syntax tree and recompute the values of each node defined by the function frame."""

    # pylint: disable=invalid-name
    # pylint: disable=missing-docstring
    # pylint: disable=too-many-public-methods

    def __init__(self, variable_lookup: List[Mapping[str, Any]]) -> None:
        """
        Initialize.

        :param variable_lookup: list of lookup tables to look-up the values of the variables, sorted by precedence
        """
        self._variable_lookup = variable_lookup

        # value assigned to each visited node
        self.recomputed_values = dict()  # type: Dict[ast.AST, Any]

    def visit_Num(self, node: ast.Num) -> Union[int, float]:
        """Recompute the value as the number at the node."""
        result = node.n

        self.recomputed_values[node] = result
        return result

    def visit_Str(self, node: ast.Str) -> str:
        """Recompute the value as the string at the node."""
        result = node.s

        self.recomputed_values[node] = result
        return result

    def visit_Bytes(self, node: ast.Bytes) -> bytes:
        """Recompute the value as the bytes at the node."""
        result = node.s

        self.recomputed_values[node] = result
        return node.s

    def visit_List(self, node: ast.List) -> List[Any]:
        """Visit the elements and assemble the results into a list."""
        if isinstance(node.ctx, ast.Store):
            raise NotImplementedError("Can not compute the value of a Store on a list")

        result = [self.visit(node=elt) for elt in node.elts]

        self.recomputed_values[node] = result
        return result

    def visit_Tuple(self, node: ast.Tuple) -> Tuple[Any, ...]:
        """Visit the elements and assemble the results into a tuple."""
        if isinstance(node.ctx, ast.Store):
            raise NotImplementedError("Can not compute the value of a Store on a tuple")

        result = tuple(self.visit(node=elt) for elt in node.elts)

        self.recomputed_values[node] = result
        return result

    def visit_Set(self, node: ast.Set) -> Set[Any]:
        """Visit the elements and assemble the results into a set."""
        result = set(self.visit(node=elt) for elt in node.elts)

        self.recomputed_values[node] = result
        return result

    def visit_Dict(self, node: ast.Dict) -> Dict[Any, Any]:
        """Visit keys and values and assemble a dictionary with the results."""
        recomputed_dict = dict()  # type: Dict[Any, Any]
        for key, val in zip(node.keys, node.values):
            recomputed_dict[self.visit(node=key)] = self.visit(node=val)

        self.recomputed_values[node] = recomputed_dict
        return recomputed_dict

    def visit_NameConstant(self, node: ast.NameConstant) -> Any:
        """Forward the node value as a result."""
        self.recomputed_values[node] = node.value
        return node.value

    def visit_Name(self, node: ast.Name) -> Any:
        """Load the variable by looking it up in the variable look-up and in the built-ins."""
        if not isinstance(node.ctx, ast.Load):
            raise NotImplementedError("Can only compute a value of Load on a name {}, but got context: {}".format(
                node.id, node.ctx))

        result = None  # type: Optional[Any]
        for lookup in self._variable_lookup:
            if node.id in lookup:
                result = lookup[node.id]
                break

        if result is None and hasattr(builtins, node.id):
            result = getattr(builtins, node.id)

        if result is None and node.id != "None":
            raise ValueError("Name not found in the variable lookup: {}".format(node.id))

        self.recomputed_values[node] = result
        return result

    def visit_Expr(self, node: ast.Expr) -> Any:
        """Visit the node's ``value``."""
        result = self.visit(node=node.value)

        self.recomputed_values[node] = result
        return result

    def visit_UnaryOp(self, node: ast.UnaryOp) -> Any:
        """Visit the node operand and apply the operation on the result."""
        if isinstance(node.op, ast.UAdd):
            result = +self.visit(node=node.operand)
        elif isinstance(node.op, ast.USub):
            result = -self.visit(node=node.operand)
        elif isinstance(node.op, ast.Not):
            result = not self.visit(node=node.operand)
        elif isinstance(node.op, ast.Invert):
            result = ~self.visit(node=node.operand)
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        """Recursively visit the left and right operand, respectively, and apply the operation on the results."""
        # pylint: disable=too-many-branches
        left = self.visit(node=node.left)
        right = self.visit(node=node.right)

        if isinstance(node.op, ast.Add):
            result = left + right
        elif isinstance(node.op, ast.Sub):
            result = left - right
        elif isinstance(node.op, ast.Mult):
            result = left * right
        elif isinstance(node.op, ast.Div):
            result = left / right
        elif isinstance(node.op, ast.FloorDiv):
            result = left // right
        elif isinstance(node.op, ast.Mod):
            result = left % right
        elif isinstance(node.op, ast.Pow):
            result = left**right
        elif isinstance(node.op, ast.LShift):
            result = left << right
        elif isinstance(node.op, ast.RShift):
            result = left >> right
        elif isinstance(node.op, ast.BitOr):
            result = left | right
        elif isinstance(node.op, ast.BitXor):
            result = left ^ right
        elif isinstance(node.op, ast.BitAnd):
            result = left & right
        elif isinstance(node.op, ast.MatMult):
            result = left @ right
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_BoolOp(self, node: ast.BoolOp) -> Any:
        """Recursively visit the operands and apply the operation on them."""
        values = [self.visit(value_node) for value_node in node.values]

        if isinstance(node.op, ast.And):
            result = functools.reduce(lambda left, right: left and right, values, True)
        elif isinstance(node.op, ast.Or):
            result = functools.reduce(lambda left, right: left or right, values, True)
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_Compare(self, node: ast.Compare) -> Any:
        """Recursively visit the comparators and apply the operations on them."""
        # pylint: disable=too-many-branches
        left = self.visit(node=node.left)

        comparators = [self.visit(node=comparator) for comparator in node.comparators]

        result = True
        for comparator, op in zip(comparators, node.ops):
            if isinstance(op, ast.Eq):
                comparison = left == comparator
            elif isinstance(op, ast.NotEq):
                comparison = left != comparator
            elif isinstance(op, ast.Lt):
                comparison = left < comparator
            elif isinstance(op, ast.LtE):
                comparison = left <= comparator
            elif isinstance(op, ast.Gt):
                comparison = left > comparator
            elif isinstance(op, ast.GtE):
                comparison = left >= comparator
            elif isinstance(op, ast.Is):
                comparison = left is comparator
            elif isinstance(op, ast.IsNot):
                comparison = left is not comparator
            elif isinstance(op, ast.In):
                comparison = left in comparator
            elif isinstance(op, ast.NotIn):
                comparison = left not in comparator
            else:
                raise NotImplementedError("Unhandled op of {}: {}".format(node, op))

            if not comparison:
                result = False
                break

            left = comparator

        self.recomputed_values[node] = result
        return result

    def visit_Call(self, node: ast.Call) -> Any:
        """Visit the function and the arguments and finally make the function call with them."""
        func = self.visit(node=node.func)

        args = []  # type: List[Any]
        for arg_node in node.args:
            if isinstance(arg_node, ast.Starred):
                args.extend(self.visit(node=arg_node))
            else:
                args.append(self.visit(node=arg_node))

        kwargs = dict()  # type: Dict[str, Any]
        for keyword in node.keywords:
            if keyword.arg is None:
                kw = self.visit(node=keyword.value)
                for key, val in kw.items():
                    kwargs[key] = val

            else:
                kwargs[keyword.arg] = self.visit(node=keyword.value)

        result = func(*args, **kwargs)

        self.recomputed_values[node] = result
        return result

    def visit_IfExp(self, node: ast.IfExp) -> Any:
        """Visit the ``test``, and depending on its outcome, the ``body`` or ``orelse``."""
        test = self.visit(node=node.test)

        if test:
            result = self.visit(node=node.body)
        else:
            result = self.visit(node=node.orelse)

        self.recomputed_values[node] = result
        return result

    def visit_Attribute(self, node: ast.Attribute) -> Any:
        """Visit the node's ``value`` and get the attribute from the result."""
        value = self.visit(node=node.value)
        if not isinstance(node.ctx, ast.Load):
            raise NotImplementedError(
                "Can only compute a value of Load on the attribute {}, but got context: {}".format(node.attr, node.ctx))

        result = getattr(value, node.attr)

        self.recomputed_values[node] = result
        return result

    def visit_Index(self, node: ast.Index) -> Any:
        """Visit the node's ``value``."""
        result = self.visit(node=node.value)

        self.recomputed_values[node] = result
        return result

    def visit_Slice(self, node: ast.Slice) -> slice:
        """Visit ``lower``, ``upper`` and ``step`` and recompute the node as a ``slice``."""
        lower = None  # type: Optional[int]
        if node.lower is not None:
            lower = self.visit(node=node.lower)

        upper = None  # type: Optional[int]
        if node.upper is not None:
            upper = self.visit(node=node.upper)

        step = None  # type: Optional[int]
        if node.step is not None:
            step = self.visit(node=node.step)

        result = slice(lower, upper, step)

        self.recomputed_values[node] = result
        return result

    def visit_ExtSlice(self, node: ast.ExtSlice) -> Tuple[Any, ...]:
        """Visit each dimension of the advanced slicing and assemble the dimensions in a tuple."""
        result = tuple(self.visit(node=dim) for dim in node.dims)

        self.recomputed_values[node] = result
        return result

    def visit_Subscript(self, node: ast.Subscript) -> Any:
        """Visit the ``slice`` and a ``value`` and get the element."""
        value = self.visit(node=node.value)
        a_slice = self.visit(node=node.slice)

        result = value[a_slice]

        self.recomputed_values[node] = result
        return result

    def visit_Lambda(self, node: ast.Lambda) -> Callable:
        """Visit the lambda's body."""
        result = self.visit(node.body)

        self.recomputed_values[node] = result
        return result

    def visit_Return(self, node: ast.Return) -> Any:  # pylint: disable=no-self-use
        """Raise an exception that this node is unexpected."""
        raise AssertionError("Unexpected return node during the re-computation: {}".format(ast.dump(node)))

    def generic_visit(self, node: ast.AST) -> None:
        """Raise an exception that this node has not been handled."""
        raise NotImplementedError("Unhandled recomputation of the node: {} {}".format(type(node), node))



"""Handle representations necessary for informative error messages."""
import ast
import inspect
import reprlib
from typing import Any, Mapping, MutableMapping, Callable, List, Dict  # pylint: disable=unused-import

import meta

import icontract.recompute


def _representable(value: Any) -> bool:
    """
    Check whether we want to represent the value in the error message on contract breach.

    We do not want to represent classes, methods, modules and functions.

    :param value: value related to an AST node
    :return: True if we want to represent it in the violation error
    """
    return not inspect.isclass(value) and not inspect.isfunction(value) and not inspect.ismethod(value) and not \
        inspect.ismodule(value) and not inspect.isbuiltin(value)


class Visitor(ast.NodeVisitor):
    """Traverse the abstract syntax tree and collect the representations of the selected nodes."""

    # pylint: disable=invalid-name
    # pylint: disable=missing-docstring

    def __init__(self, recomputed_values: Mapping[ast.AST, Any], variable_lookup: List[Mapping[str, Any]]) -> None:
        """
        Initialize.

        :param recomputed_values: AST node of a condition function -> value associated with the node
        :param variable_lookup:
            list of lookup tables to look-up the values of the variables, sorted by precedence.
            The visitor needs it here to check whether we overrode a built-in variable (like ``id``).

        """
        self._recomputed_values = recomputed_values
        self._variable_lookup = variable_lookup
        self.reprs = dict()  # type: MutableMapping[str, str]

    def visit_Name(self, node: ast.Name) -> None:
        """
        Resolve the name from the variable look-up and the built-ins.

        Due to possible branching (e.g., If-expressions), some nodes might lack the recomputed values. These nodes
        are ignored.
        """
        if node in self._recomputed_values:
            value = self._recomputed_values[node]

            # Check if it is a non-built-in
            is_builtin = True
            for lookup in self._variable_lookup:
                if node.id in lookup:
                    is_builtin = False
                    break

            if not is_builtin and _representable(value=value):
                text = str(meta.dump_python_source(node)).strip()  # type: ignore
                self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        """Represent the attribute by dumping its source code."""
        value = self._recomputed_values[node]

        if _representable(value=value):
            text = str(meta.dump_python_source(node)).strip()  # type: ignore
            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_Call(self, node: ast.Call) -> None:
        """Represent the call by dumping its source code."""
        value = self._recomputed_values[node]

        # pylint: disable=no-member
        text = str(meta.dump_python_source(node)).strip()  # type: ignore
        self.reprs[text] = value

        self.generic_visit(node=node)


def _is_lambda(condition: Callable[..., bool]) -> bool:
    """
    Check whether the condition is a lambda function.

    >>> def some_func()->bool: return True
    >>> _is_lambda(some_func)
    False

    >>> lmbd = lambda x: x > 0
    >>> _is_lambda(lmbd)
    True

    :param condition: condition function of a contract
    :return: True if condition is defined as lambda function
    """
    return condition.__name__ == "<lambda>"


def repr_values(condition: Callable[..., bool], condition_kwargs: Mapping[str, Any], a_repr: reprlib.Repr) -> List[str]:
    """
    Represent function arguments and frame values in the error message on contract breach.

    :param condition: contract condition function
    :param condition_kwargs: condition arguments
    :param a_repr: representation instance that defines how the values are represented.
    :return: list of value representations
    """
    # pylint: disable=too-many-locals
    reprs = dict()  # type: MutableMapping[str, Any]

    if _is_lambda(condition=condition):
        # pylint: disable=no-member

        root_node = meta.decompiler.decompile_func(condition)
        assert isinstance(root_node, ast.Lambda), \
            "Expected the node of the decompiled condition function to be a Lambda, but got: {}".format(
                ast.dump(root_node))

        assert isinstance(root_node.body, ast.Return), \
            "Expected the body of the de-compiled lambda function to be a Return, but got: {}".format(
                ast.dump(root_node))

        lambda_expression = root_node.body.value

        # pylint: enable=no-member

        # Collect the variable lookup of the condition function:
        # globals, closure, defaults, kwdefaults, arguments
        variable_lookup = []  # type: List[Dict[str, Any]]

        if condition.__globals__ is not None:
            variable_lookup.append(condition.__globals__)

        closure_dict = dict()  # type: Dict[str, Any]

        if condition.__closure__ is not None:
            closure_cells = condition.__closure__
            freevars = condition.__code__.co_freevars

            assert len(closure_cells) == len(freevars), \
                "Number of closure cells of a condition function ({}) == number of free vars ({})".format(
                    len(closure_cells), len(freevars))

            for cell, freevar in zip(closure_cells, freevars):
                closure_dict[freevar] = cell.cell_contents

        variable_lookup.append(closure_dict)

        if condition.__kwdefaults__ is not None:
            variable_lookup.append(condition.__kwdefaults__)

        variable_lookup.append(condition_kwargs)

        recompute_visitor = icontract.recompute.Visitor(variable_lookup=variable_lookup)

        recompute_visitor.visit(node=lambda_expression)
        recomputed_values = recompute_visitor.recomputed_values

        repr_visitor = Visitor(recomputed_values=recomputed_values, variable_lookup=variable_lookup)
        repr_visitor.visit(node=lambda_expression)

        reprs = repr_visitor.reprs
    else:
        for key, val in condition_kwargs.items():
            if _representable(value=val):
                reprs[key] = val

    parts = []  # type: List[str]
    for key in sorted(reprs.keys()):
        parts.append('{} was {}'.format(key, a_repr.repr(reprs[key])))

    return parts


def condition_as_text(condition: Callable[..., bool]) -> str:
    """
    Decompile the condition function and generate the text based on its abstract syntax tree.

    The decompilation and generation is necessary since it is hard to otherwise parse the output of inspect.getsource()
    on decorator functions.

    :param condition: contract condition function
    :return: string representation of the condition function
    """
    if not _is_lambda(condition):
        return condition.__name__

    # pylint: disable=no-member

    lambda_node = meta.decompiler.decompile_func(condition)
    assert isinstance(lambda_node, ast.Lambda), \
        "Expected the root node of the de-compiled condition to be a Lambda, but got: {}".format(
            ast.dump(lambda_node))

    assert isinstance(lambda_node.body, ast.Return), \
        "Expected the body of the de-compiled condition to be a Return, but got: {}".format(
            ast.dump(lambda_node.body))

    body_txt = str(meta.dump_python_source(lambda_node.body.value)).strip()  # type: ignore

    # pylint: enable=no-member

    # Strip enclosing brackets from the body
    if body_txt.startswith("(") and body_txt.endswith(")"):
        body_txt = body_txt[1:-1]

    return body_txt
